#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import zmq

# HTTP Scheme
HTTP_SCHEME = "http"
HTTPS_SCHEME = "https"

# HTTP Port
HTTP_PORT = 80
HTTPS_PORT = 443

# HTTP Request Methods
HTTP_METHOD_GET = "get"
HTTP_METHOD_HEAD = "head"
HTTP_METHOD_POST = "post"
HTTP_METHOD_PUT = "put"
HTTP_METHOD_DELETE = "delete"
HTTP_METHOD_CONNECT = "connect"
HTTP_METHOD_OPTIONS = "options"
HTTP_METHOD_TRACE = "trace"
HTTP_METHOD_PATCH = "patch"

# HTTP Headers
HTTP_HEADER_ACCEPT = "Accept"
HTTP_HEADER_ACCEPT_CH = "Accept-CH"
HTTP_HEADER_ACCEPT_CH_LIFETIME = "Accept-CH-Lifetime"
HTTP_HEADER_ACCEPT_CHARSET = "Accept-Charset"
HTTP_HEADER_ACCEPT_ENCODING = "Accept-Encoding"
HTTP_HEADER_ACCEPT_LANGUAGE = "Accept-Language"
HTTP_HEADER_ACCEPT_PATCH = "Accept-Patch"
HTTP_HEADER_ACCEPT_POST = "Accept-Post"
HTTP_HEADER_ACCEPT_RANGES = "Accept-Ranges"
HTTP_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS = "Access-Control-Allow-Credentials"
HTTP_HEADER_ACCESS_CONTROL_ALLOW_HEADERS = "Access-Control-Allow-Headers"
HTTP_HEADER_ACCESS_CONTROL_ALLOW_METHODS = "Access-Control-Allow-Methods"
HTTP_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN = "Access-Control-Allow-Origin"
HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS = "Access-Control-Expose-Headers"
HTTP_HEADER_ACCESS_CONTROL_MAX_AGE = "Access-Control-Max-Age"
HTTP_HEADER_ACCESS_CONTROL_REQUEST_HEADERS = "Access-Control-Request-Headers"
HTTP_HEADER_ACCESS_CONTROL_REQUEST_METHOD = "Access-Control-Request-Method"
HTTP_HEADER_AGE = "Age"
HTTP_HEADER_ALLOW = "Allow"
HTTP_HEADER_ALT_SVC = "Alt-Svc"
HTTP_HEADER_AUTHORIZATION = "Authorization"
HTTP_HEADER_CACHE_CONTROL = "Cache-Control"
HTTP_HEADER_CLEAR_SITE_DATA = "Clear-Site-Data"
HTTP_HEADER_CONNECTION = "Connection"
HTTP_HEADER_CONTENT_DISPOSITION = "Content-Disposition"
HTTP_HEADER_CONTENT_DPR = "Content-DPR"
HTTP_HEADER_CONTENT_ENCODING = "Content-Encoding"
HTTP_HEADER_CONTENT_LANGUAGE = "Content-Language"
HTTP_HEADER_CONTENT_LENGTH = "Content-Length"
HTTP_HEADER_CONTENT_LOCATION = "Content-Location"
HTTP_HEADER_CONTENT_RANGE = "Content-Range"
HTTP_HEADER_CONTENT_SECURITY_POLICY = "Content-Security-Policy"
HTTP_HEADER_CONTENT_SECURITY_POLICY_REPORT_ONLY = "Content-Security-Policy-Report-Only"
HTTP_HEADER_CONTENT_TYPE = "Content-Type"
HTTP_HEADER_COOKIE = "Cookie"
HTTP_HEADER_CRITICAL_CH = "Critical-CH"
HTTP_HEADER_CROSS_ORIGIN_EMBEDDER_POLICY = "Cross-Origin-Embedder-Policy"
HTTP_HEADER_CROSS_ORIGIN_OPENER_POLICY = "Cross-Origin-Opener-Policy"
HTTP_HEADER_CROSS_ORIGIN_RESOURCE_POLICY = "Cross-Origin-Resource-Policy"
HTTP_HEADER_DATE = "Date"
HTTP_HEADER_DEVICE_MEMORY = "Device-Memory"
HTTP_HEADER_DIGEST = "Digest"
HTTP_HEADER_DNT = "DNT"
HTTP_HEADER_DOWNLINK = "Downlink"
HTTP_HEADER_DPR = "DPR"
HTTP_HEADER_EARLY_DATA = "Early-Data"
HTTP_HEADER_ECT = "ECT"
HTTP_HEADER_ETAG = "ETag"
HTTP_HEADER_EXPECT = "Expect"
HTTP_HEADER_EXPECT_CT = "Expect-CT"
HTTP_HEADER_EXPIRES = "Expires"
HTTP_HEADER_FORWARDED = "Forwarded"
HTTP_HEADER_FROM = "From"
HTTP_HEADER_HOST = "Host"
HTTP_HEADER_IF_MATCH = "If-Match"
HTTP_HEADER_IF_MODIFIED_SINCE = "If-Modified-Since"
HTTP_HEADER_IF_NONE_MATCH = "If-None-Match"
HTTP_HEADER_IF_RANGE = "If-Range"
HTTP_HEADER_IF_UNMODIFIED_SINCE = "If-Unmodified-Since"
HTTP_HEADER_KEEP_ALIVE = "Keep-Alive"
HTTP_HEADER_LARGE_ALLOCATION = "Large-Allocation"
HTTP_HEADER_LAST_MODIFIED = "Last-Modified"
HTTP_HEADER_LINK = "Link"
HTTP_HEADER_LOCATION = "Location"
HTTP_HEADER_MAX_FORWARDS = "Max-Forwards"
HTTP_HEADER_NEL = "NEL"
HTTP_HEADER_ORIGIN = "Origin"
HTTP_HEADER_PERMISSIONS_POLICY = "Permissions-Policy"
HTTP_HEADER_PRAGMA = "Pragma"
HTTP_HEADER_PROXY_AUTHENTICATE = "Proxy-Authenticate"
HTTP_HEADER_PROXY_AUTHORIZATION = "Proxy-Authorization"
HTTP_HEADER_RANGE = "Range"
HTTP_HEADER_REFERER = "Referer"
HTTP_HEADER_REFERRER_POLICY = "Referrer-Policy"
HTTP_HEADER_RETRY_AFTER = "Retry-After"
HTTP_HEADER_RTT = "RTT"
HTTP_HEADER_SAVE_DATA = "Save-Data"
HTTP_HEADER_SEC_CH_PREFERS_REDUCED_MOTION = "Sec-CH-Prefers-Reduced-Motion"
HTTP_HEADER_SEC_CH_UA = "Sec-CH-UA"
HTTP_HEADER_SEC_CH_UA_ARCH = "Sec-CH-UA-Arch"
HTTP_HEADER_SEC_CH_UA_BITNESS = "Sec-CH-UA-Bitness"
HTTP_HEADER_SEC_CH_UA_FULL_VERSION = "Sec-CH-UA-Full-Version"
HTTP_HEADER_SEC_CH_UA_FULL_VERSION_LIST = "Sec-CH-UA-Full-Version-List"
HTTP_HEADER_SEC_CH_UA_MOBILE = "Sec-CH-UA-Mobile"
HTTP_HEADER_SEC_CH_UA_PLATFORM = "Sec-CH-UA-Platform"
HTTP_HEADER_SEC_CH_UA_PLATFORM_VERSION = "Sec-CH-UA-Platform-Version"
HTTP_HEADER_SEC_FETCH_DEST = "Sec-Fetch-Dest"
HTTP_HEADER_SEC_FETCH_MODE = "Sec-Fetch-Mode"
HTTP_HEADER_SEC_FETCH_SITE = "Sec-Fetch-Site"
HTTP_HEADER_SEC_FETCH_USER = "Sec-Fetch-User"
HTTP_HEADER_SEC_GPC = "Sec-GPC"
HTTP_HEADER_SEC_WEBSOCKET_ACCEPT = "Sec-WebSocket-Accept"
HTTP_HEADER_SERVER = "Server"
HTTP_HEADER_SERVER_TIMING = "Server-Timing"
HTTP_HEADER_SERVICE_WORKER_NAVIGATION_PRELOAD = "Service-Worker-Navigation-Preload"
HTTP_HEADER_SET_COOKIE = "Set-Cookie"
HTTP_HEADER_SOURCEMAP = "SourceMap"
HTTP_HEADER_STRICT_TRANSPORT_SECURITY = "Strict-Transport-Security"
HTTP_HEADER_TE = "TE"
HTTP_HEADER_TIMING_ALLOW_ORIGIN = "Timing-Allow-Origin"
HTTP_HEADER_TK = "Tk"
HTTP_HEADER_TRAILER = "Trailer"
HTTP_HEADER_TRANSFER_ENCODING = "Transfer-Encoding"
HTTP_HEADER_UPGRADE = "Upgrade"
HTTP_HEADER_UPGRADE_INSECURE_REQUESTS = "Upgrade-Insecure-Requests"
HTTP_HEADER_USER_AGENT = "User-Agent"
HTTP_HEADER_VARY = "Vary"
HTTP_HEADER_VIA = "Via"
HTTP_HEADER_VIEWPORT_WIDTH = "Viewport-Width"
HTTP_HEADER_WANT_DIGEST = "Want-Digest"
HTTP_HEADER_WARNING = "Warning"
HTTP_HEADER_WIDTH = "Width"
HTTP_HEADER_WWW_AUTHENTICATE = "WWW-Authenticate"
HTTP_HEADER_X_CONTENT_TYPE_OPTIONS = "X-Content-Type-Options"
HTTP_HEADER_X_DNS_PREFETCH_CONTROL = "X-DNS-Prefetch-Control"
HTTP_HEADER_X_FORWARDED_FOR = "X-Forwarded-For"
HTTP_HEADER_X_FORWARDED_HOST = "X-Forwarded-Host"
HTTP_HEADER_X_FORWARDED_PROTO = "X-Forwarded-Proto"
HTTP_HEADER_X_FRAME_OPTIONS = "X-Frame-Options"
HTTP_HEADER_X_XSS_PROTECTION = "X-XSS-Protection"

HTTP_HEADERS = [
                HTTP_HEADER_ACCEPT,
                HTTP_HEADER_ACCEPT_CH,
                HTTP_HEADER_ACCEPT_CH_LIFETIME,
                HTTP_HEADER_ACCEPT_CHARSET,
                HTTP_HEADER_ACCEPT_ENCODING,
                HTTP_HEADER_ACCEPT_LANGUAGE,
                HTTP_HEADER_ACCEPT_PATCH,
                HTTP_HEADER_ACCEPT_POST,
                HTTP_HEADER_ACCEPT_RANGES,
                HTTP_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS,
                HTTP_HEADER_ACCESS_CONTROL_ALLOW_HEADERS,
                HTTP_HEADER_ACCESS_CONTROL_ALLOW_METHODS,
                HTTP_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN,
                HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
                HTTP_HEADER_ACCESS_CONTROL_MAX_AGE,
                HTTP_HEADER_ACCESS_CONTROL_REQUEST_HEADERS,
                HTTP_HEADER_ACCESS_CONTROL_REQUEST_METHOD,
                HTTP_HEADER_AGE,
                HTTP_HEADER_ALLOW,
                HTTP_HEADER_ALT_SVC,
                HTTP_HEADER_AUTHORIZATION,
                HTTP_HEADER_CACHE_CONTROL,
                HTTP_HEADER_CLEAR_SITE_DATA,
                HTTP_HEADER_CONNECTION,
                HTTP_HEADER_CONTENT_DISPOSITION,
                HTTP_HEADER_CONTENT_DPR,
                HTTP_HEADER_CONTENT_ENCODING,
                HTTP_HEADER_CONTENT_LANGUAGE,
                HTTP_HEADER_CONTENT_LENGTH,
                HTTP_HEADER_CONTENT_LOCATION,
                HTTP_HEADER_CONTENT_RANGE,
                HTTP_HEADER_CONTENT_SECURITY_POLICY,
                HTTP_HEADER_CONTENT_SECURITY_POLICY_REPORT_ONLY,
                HTTP_HEADER_CONTENT_TYPE,
                HTTP_HEADER_COOKIE,
                HTTP_HEADER_CRITICAL_CH,
                HTTP_HEADER_CROSS_ORIGIN_EMBEDDER_POLICY,
                HTTP_HEADER_CROSS_ORIGIN_OPENER_POLICY,
                HTTP_HEADER_CROSS_ORIGIN_RESOURCE_POLICY,
                HTTP_HEADER_DATE,
                HTTP_HEADER_DEVICE_MEMORY,
                HTTP_HEADER_DIGEST,
                HTTP_HEADER_DNT,
                HTTP_HEADER_DOWNLINK,
                HTTP_HEADER_DPR,
                HTTP_HEADER_EARLY_DATA,
                HTTP_HEADER_ECT,
                HTTP_HEADER_ETAG,
                HTTP_HEADER_EXPECT,
                HTTP_HEADER_EXPECT_CT,
                HTTP_HEADER_EXPIRES,
                HTTP_HEADER_FORWARDED,
                HTTP_HEADER_FROM,
                HTTP_HEADER_HOST,
                HTTP_HEADER_IF_MATCH,
                HTTP_HEADER_IF_MODIFIED_SINCE,
                HTTP_HEADER_IF_NONE_MATCH,
                HTTP_HEADER_IF_RANGE,
                HTTP_HEADER_IF_UNMODIFIED_SINCE,
                HTTP_HEADER_KEEP_ALIVE,
                HTTP_HEADER_LARGE_ALLOCATION,
                HTTP_HEADER_LAST_MODIFIED,
                HTTP_HEADER_LINK,
                HTTP_HEADER_LOCATION,
                HTTP_HEADER_MAX_FORWARDS,
                HTTP_HEADER_NEL,
                HTTP_HEADER_ORIGIN,
                HTTP_HEADER_PERMISSIONS_POLICY,
                HTTP_HEADER_PRAGMA,
                HTTP_HEADER_PROXY_AUTHENTICATE,
                HTTP_HEADER_PROXY_AUTHORIZATION,
                HTTP_HEADER_RANGE,
                HTTP_HEADER_REFERER,
                HTTP_HEADER_REFERRER_POLICY,
                HTTP_HEADER_RETRY_AFTER,
                HTTP_HEADER_RTT,
                HTTP_HEADER_SAVE_DATA,
                HTTP_HEADER_SEC_CH_PREFERS_REDUCED_MOTION,
                HTTP_HEADER_SEC_CH_UA,
                HTTP_HEADER_SEC_CH_UA_ARCH,
                HTTP_HEADER_SEC_CH_UA_BITNESS,
                HTTP_HEADER_SEC_CH_UA_FULL_VERSION,
                HTTP_HEADER_SEC_CH_UA_FULL_VERSION_LIST,
                HTTP_HEADER_SEC_CH_UA_MOBILE,
                HTTP_HEADER_SEC_CH_UA_PLATFORM,
                HTTP_HEADER_SEC_CH_UA_PLATFORM_VERSION,
                HTTP_HEADER_SEC_FETCH_DEST,
                HTTP_HEADER_SEC_FETCH_MODE,
                HTTP_HEADER_SEC_FETCH_SITE,
                HTTP_HEADER_SEC_FETCH_USER,
                HTTP_HEADER_SEC_GPC,
                HTTP_HEADER_SEC_WEBSOCKET_ACCEPT,
                HTTP_HEADER_SERVER,
                HTTP_HEADER_SERVER_TIMING,
                HTTP_HEADER_SERVICE_WORKER_NAVIGATION_PRELOAD,
                HTTP_HEADER_SET_COOKIE,
                HTTP_HEADER_SOURCEMAP,
                HTTP_HEADER_STRICT_TRANSPORT_SECURITY,
                HTTP_HEADER_TE,
                HTTP_HEADER_TIMING_ALLOW_ORIGIN,
                HTTP_HEADER_TK,
                HTTP_HEADER_TRAILER,
                HTTP_HEADER_TRANSFER_ENCODING,
                HTTP_HEADER_UPGRADE,
                HTTP_HEADER_UPGRADE_INSECURE_REQUESTS,
                HTTP_HEADER_USER_AGENT,
                HTTP_HEADER_VARY,
                HTTP_HEADER_VIA,
                HTTP_HEADER_VIEWPORT_WIDTH,
                HTTP_HEADER_WANT_DIGEST,
                HTTP_HEADER_WARNING,
                HTTP_HEADER_WIDTH,
                HTTP_HEADER_WWW_AUTHENTICATE,
                HTTP_HEADER_X_CONTENT_TYPE_OPTIONS,
                HTTP_HEADER_X_DNS_PREFETCH_CONTROL,
                HTTP_HEADER_X_FORWARDED_FOR,
                HTTP_HEADER_X_FORWARDED_HOST,
                HTTP_HEADER_X_FORWARDED_PROTO,
                HTTP_HEADER_X_FRAME_OPTIONS,
                HTTP_HEADER_X_XSS_PROTECTION
               ]

# ZeroMq Types
ZMQ_TYPE_PAIR = "PAIR"
ZMQ_TYPE_PUB = "PUB"
ZMQ_TYPE_SUB = "SUB"
ZMQ_TYPE_REQ = "REQ"
ZMQ_TYPE_REP = "REP"
ZMQ_TYPE_DEALER = "DEALER"
ZMQ_TYPE_ROUTER = "ROUTER"
ZMQ_TYPE_PULL = "PULL"
ZMQ_TYPE_PUSH = "PUSH"
ZMQ_TYPE_XPUB = "XPUB"
ZMQ_TYPE_XSUB = "XSUB"
ZMQ_TYPE_STREAM = "STREAM"
ZMQ_TYPE_XREQ = "XREQ"
ZMQ_TYPE_XREP = "XREP"
ZMQ_TYPE_SERVER = "SERVER"
ZMQ_TYPE_CLIENT = "CLIENT"
ZMQ_TYPE_RADIO = "RADIO"
ZMQ_TYPE_DISH = "DISH"
ZMQ_TYPE_GATHER = "GATHER"
ZMQ_TYPE_SCATTER = "SCATTER"
ZMQ_TYPE_DGRAM = "DGRAM"
ZMQ_TYPE_PEER = "PEER"
ZMQ_TYPE_CHANNEL = "CHANNEL"

ZMQ_TYPES = {
             ZMQ_TYPE_PAIR: zmq.PAIR,
             ZMQ_TYPE_PUB: zmq.PUB,
             ZMQ_TYPE_SUB: zmq.SUB,
             ZMQ_TYPE_REQ: zmq.REQ,
             ZMQ_TYPE_REP: zmq.REP,
             ZMQ_TYPE_DEALER: zmq.DEALER,
             ZMQ_TYPE_ROUTER: zmq.ROUTER,
             ZMQ_TYPE_PULL: zmq.PULL,
             ZMQ_TYPE_PUSH: zmq.PUSH,
             ZMQ_TYPE_XPUB: zmq.XPUB,
             ZMQ_TYPE_XSUB: zmq.XSUB,
             ZMQ_TYPE_STREAM: zmq.STREAM,
             ZMQ_TYPE_XREQ: zmq.XREQ,
             ZMQ_TYPE_XREP: zmq.XREP,
             ZMQ_TYPE_SERVER: zmq.SERVER,
             ZMQ_TYPE_RADIO: zmq.RADIO,
             ZMQ_TYPE_DISH: zmq.DISH,
             ZMQ_TYPE_GATHER: zmq.GATHER,
             ZMQ_TYPE_SCATTER: zmq.SCATTER,
             ZMQ_TYPE_DGRAM: zmq.DGRAM,
             ZMQ_TYPE_PEER: zmq.PEER,
             ZMQ_TYPE_CHANNEL: zmq.CHANNEL
            }

# ZeroMq Socket Options
ZMQ_HWM = "HWM"
ZMQ_AFFINITY = "AFFINITY"
ZMQ_ROUTING_ID = "ROUTING_ID"
ZMQ_SUBSCRIBE = "SUBSCRIBE"
ZMQ_UNSUBSCRIBE = "UNSUBSCRIBE"
ZMQ_RATE = "RATE"
ZMQ_RECOVERY_IVL = "RECOVERY_IVL"
ZMQ_SNDBUF = "SNDBUF"
ZMQ_RCVBUF = "RCVBUF"
ZMQ_RCVMORE = "RCVMORE"
ZMQ_FD = "FD"
ZMQ_EVENTS = "EVENTS"
ZMQ_TYPE = "TYPE"
ZMQ_LINGER = "LINGER"
ZMQ_RECONNECT_IVL = "RECONNECT_IVL"
ZMQ_BACKLOG = "BACKLOG"
ZMQ_RECONNECT_IVL_MAX = "RECONNECT_IVL_MAX"
ZMQ_MAXMSGSIZE = "MAXMSGSIZE"
ZMQ_SNDHWM = "SNDHWM"
ZMQ_RCVHWM = "RCVHWM"
ZMQ_MULTICAST_HOPS = "MULTICAST_HOPS"
ZMQ_RCVTIMEO = "RCVTIMEO"
ZMQ_SNDTIMEO = "SNDTIMEO"
ZMQ_LAST_ENDPOINT = "LAST_ENDPOINT"
ZMQ_ROUTER_MANDATORY = "ROUTER_MANDATORY"
ZMQ_TCP_KEEPALIVE = "TCP_KEEPALIVE"
ZMQ_TCP_KEEPALIVE_CNT = "TCP_KEEPALIVE_CNT"
ZMQ_TCP_KEEPALIVE_IDLE = "TCP_KEEPALIVE_IDLE"
ZMQ_TCP_KEEPALIVE_INTVL = "TCP_KEEPALIVE_INTVL"
ZMQ_IMMEDIATE = "IMMEDIATE"
ZMQ_XPUB_VERBOSE = "XPUB_VERBOSE"
ZMQ_ROUTER_RAW = "ROUTER_RAW"
ZMQ_IPV6 = "IPV6"
ZMQ_MECHANISM = "MECHANISM"
ZMQ_PLAIN_SERVER = "PLAIN_SERVER"
ZMQ_PLAIN_USERNAME = "PLAIN_USERNAME"
ZMQ_PLAIN_PASSWORD = "PLAIN_PASSWORD"
ZMQ_CURVE_SERVER = "CURVE_SERVER"
ZMQ_CURVE_PUBLICKEY = "CURVE_PUBLICKEY"
ZMQ_CURVE_SECRETKEY = "CURVE_SECRETKEY"
ZMQ_CURVE_SERVERKEY = "CURVE_SERVERKEY"
ZMQ_PROBE_ROUTER = "PROBE_ROUTER"
ZMQ_REQ_CORRELATE = "REQ_CORRELATE"
ZMQ_REQ_RELAXED = "REQ_RELAXED"
ZMQ_CONFLATE = "CONFLATE"
ZMQ_ZAP_DOMAIN = "ZAP_DOMAIN"
ZMQ_ROUTER_HANDOVER = "ROUTER_HANDOVER"
ZMQ_TOS = "TOS"
ZMQ_CONNECT_ROUTING_ID = "CONNECT_ROUTING_ID"
ZMQ_GSSAPI_SERVER = "GSSAPI_SERVER"
ZMQ_GSSAPI_PRINCIPAL = "GSSAPI_PRINCIPAL"
ZMQ_GSSAPI_SERVICE_PRINCIPAL = "GSSAPI_SERVICE_PRINCIPAL"
ZMQ_GSSAPI_PLAINTEXT = "GSSAPI_PLAINTEXT"
ZMQ_HANDSHAKE_IVL = "HANDSHAKE_IVL"
ZMQ_SOCKS_PROXY = "SOCKS_PROXY"
ZMQ_XPUB_NODROP = "XPUB_NODROP"
ZMQ_BLOCKY = "BLOCKY"
ZMQ_XPUB_MANUAL = "XPUB_MANUAL"
ZMQ_XPUB_WELCOME_MSG = "XPUB_WELCOME_MSG"
ZMQ_STREAM_NOTIFY = "STREAM_NOTIFY"
ZMQ_INVERT_MATCHING = "INVERT_MATCHING"
ZMQ_HEARTBEAT_IVL = "HEARTBEAT_IVL"
ZMQ_HEARTBEAT_TTL = "HEARTBEAT_TTL"
ZMQ_HEARTBEAT_TIMEOUT = "HEARTBEAT_TIMEOUT"
ZMQ_XPUB_VERBOSER = "XPUB_VERBOSER"
ZMQ_CONNECT_TIMEOUT = "CONNECT_TIMEOUT"
ZMQ_TCP_MAXRT = "TCP_MAXRT"
ZMQ_THREAD_SAFE = "THREAD_SAFE"
ZMQ_MULTICAST_MAXTPDU = "MULTICAST_MAXTPDU"
ZMQ_VMCI_BUFFER_SIZE = "VMCI_BUFFER_SIZE"
ZMQ_VMCI_BUFFER_MIN_SIZE = "VMCI_BUFFER_MIN_SIZE"
ZMQ_VMCI_BUFFER_MAX_SIZE = "VMCI_BUFFER_MAX_SIZE"
ZMQ_VMCI_CONNECT_TIMEOUT = "VMCI_CONNECT_TIMEOUT"
ZMQ_USE_FD = "USE_FD"
ZMQ_GSSAPI_PRINCIPAL_NAMETYPE = "GSSAPI_PRINCIPAL_NAMETYPE"
ZMQ_GSSAPI_SERVICE_PRINCIPAL_NAMETYPE = "GSSAPI_SERVICE_PRINCIPAL_NAMETYPE"
ZMQ_BINDTODEVICE = "BINDTODEVICE"
ZMQ_IDENTITY = "IDENTITY"
ZMQ_CONNECT_RID = "CONNECT_RID"
ZMQ_TCP_ACCEPT_FILTER = "TCP_ACCEPT_FILTER"
ZMQ_IPC_FILTER_PID = "IPC_FILTER_PID"
ZMQ_IPC_FILTER_UID = "IPC_FILTER_UID"
ZMQ_IPC_FILTER_GID = "IPC_FILTER_GID"
ZMQ_IPV4ONLY = "IPV4ONLY"
ZMQ_DELAY_ATTACH_ON_CONNECT = "DELAY_ATTACH_ON_CONNECT"
ZMQ_FAIL_UNROUTABLE = "FAIL_UNROUTABLE"
ZMQ_ROUTER_BEHAVIOR = "ROUTER_BEHAVIOR"
ZMQ_ZAP_ENFORCE_DOMAIN = "ZAP_ENFORCE_DOMAIN"
ZMQ_LOOPBACK_FASTPATH = "LOOPBACK_FASTPATH"
ZMQ_METADATA = "METADATA"
ZMQ_MULTICAST_LOOP = "MULTICAST_LOOP"
ZMQ_ROUTER_NOTIFY = "ROUTER_NOTIFY"
ZMQ_XPUB_MANUAL_LAST_VALUE = "XPUB_MANUAL_LAST_VALUE"
ZMQ_SOCKS_USERNAME = "SOCKS_USERNAME"
ZMQ_SOCKS_PASSWORD = "SOCKS_PASSWORD"
ZMQ_IN_BATCH_SIZE = "IN_BATCH_SIZE"
ZMQ_OUT_BATCH_SIZE = "OUT_BATCH_SIZE"
ZMQ_WSS_KEY_PEM = "WSS_KEY_PEM"
ZMQ_WSS_CERT_PEM = "WSS_CERT_PEM"
ZMQ_WSS_TRUST_PEM = "WSS_TRUST_PEM"
ZMQ_WSS_HOSTNAME = "WSS_HOSTNAME"
ZMQ_WSS_TRUST_SYSTEM = "WSS_TRUST_SYSTEM"
ZMQ_ONLY_FIRST_SUBSCRIBE = "ONLY_FIRST_SUBSCRIBE"
ZMQ_RECONNECT_STOP = "RECONNECT_STOP"
ZMQ_HELLO_MSG = "HELLO_MSG"
ZMQ_DISCONNECT_MSG = "DISCONNECT_MSG"
ZMQ_PRIORITY = "PRIORITY"

ZMQ_SOCKET_OPTIONS = {
                      ZMQ_HWM: zmq.HWM,
                      ZMQ_AFFINITY: zmq.AFFINITY,
                      ZMQ_ROUTING_ID: zmq.ROUTING_ID,
                      ZMQ_SUBSCRIBE: zmq.SUBSCRIBE,
                      ZMQ_UNSUBSCRIBE: zmq.UNSUBSCRIBE,
                      ZMQ_RATE: zmq.RATE,
                      ZMQ_RECOVERY_IVL: zmq.RECOVERY_IVL,
                      ZMQ_SNDBUF: zmq.SNDBUF,
                      ZMQ_RCVBUF: zmq.RCVBUF,
                      ZMQ_RCVMORE: zmq.RCVMORE,
                      ZMQ_FD: zmq.FD,
                      ZMQ_EVENTS: zmq.EVENTS,
                      ZMQ_TYPE: zmq.TYPE,
                      ZMQ_LINGER: zmq.LINGER,
                      ZMQ_RECONNECT_IVL: zmq.RECONNECT_IVL,
                      ZMQ_BACKLOG: zmq.BACKLOG,
                      ZMQ_RECONNECT_IVL_MAX: zmq.RECONNECT_IVL_MAX,
                      ZMQ_MAXMSGSIZE: zmq.MAXMSGSIZE,
                      ZMQ_SNDHWM: zmq.SNDHWM,
                      ZMQ_RCVHWM: zmq.RCVHWM,
                      ZMQ_MULTICAST_HOPS: zmq.MULTICAST_HOPS,
                      ZMQ_RCVTIMEO: zmq.RCVTIMEO,
                      ZMQ_SNDTIMEO: zmq.SNDTIMEO,
                      ZMQ_LAST_ENDPOINT: zmq.LAST_ENDPOINT,
                      ZMQ_ROUTER_MANDATORY: zmq.ROUTER_MANDATORY,
                      ZMQ_TCP_KEEPALIVE: zmq.TCP_KEEPALIVE,
                      ZMQ_TCP_KEEPALIVE_CNT: zmq.TCP_KEEPALIVE_CNT,
                      ZMQ_TCP_KEEPALIVE_IDLE: zmq.TCP_KEEPALIVE_IDLE,
                      ZMQ_TCP_KEEPALIVE_INTVL: zmq.TCP_KEEPALIVE_INTVL,
                      ZMQ_IMMEDIATE: zmq.IMMEDIATE,
                      ZMQ_XPUB_VERBOSE: zmq.XPUB_VERBOSE,
                      ZMQ_ROUTER_RAW: zmq.ROUTER_RAW,
                      ZMQ_IPV6: zmq.IPV6,
                      ZMQ_MECHANISM: zmq.MECHANISM,
                      ZMQ_PLAIN_SERVER: zmq.PLAIN_SERVER,
                      ZMQ_PLAIN_USERNAME: zmq.PLAIN_USERNAME,
                      ZMQ_PLAIN_PASSWORD: zmq.PLAIN_PASSWORD,
                      ZMQ_CURVE_SERVER: zmq.CURVE_SERVER,
                      ZMQ_CURVE_PUBLICKEY: zmq.CURVE_PUBLICKEY,
                      ZMQ_CURVE_SECRETKEY: zmq.CURVE_SECRETKEY,
                      ZMQ_CURVE_SERVERKEY: zmq.CURVE_SERVERKEY,
                      ZMQ_PROBE_ROUTER: zmq.PROBE_ROUTER,
                      ZMQ_REQ_CORRELATE: zmq.REQ_CORRELATE,
                      ZMQ_REQ_RELAXED: zmq.REQ_RELAXED,
                      ZMQ_CONFLATE: zmq.CONFLATE,
                      ZMQ_ZAP_DOMAIN: zmq.ZAP_DOMAIN,
                      ZMQ_ROUTER_HANDOVER: zmq.ROUTER_HANDOVER,
                      ZMQ_TOS: zmq.TOS,
                      ZMQ_CONNECT_ROUTING_ID: zmq.CONNECT_ROUTING_ID,
                      ZMQ_GSSAPI_SERVER: zmq.GSSAPI_SERVER,
                      ZMQ_GSSAPI_PRINCIPAL: zmq.GSSAPI_PRINCIPAL,
                      ZMQ_GSSAPI_SERVICE_PRINCIPAL: zmq.GSSAPI_SERVICE_PRINCIPAL,
                      ZMQ_GSSAPI_PLAINTEXT: zmq.GSSAPI_PLAINTEXT,
                      ZMQ_HANDSHAKE_IVL: zmq.HANDSHAKE_IVL,
                      ZMQ_SOCKS_PROXY: zmq.SOCKS_PROXY,
                      ZMQ_XPUB_NODROP: zmq.XPUB_NODROP,
                      ZMQ_BLOCKY: zmq.BLOCKY,
                      ZMQ_XPUB_MANUAL: zmq.XPUB_MANUAL,
                      ZMQ_XPUB_WELCOME_MSG: zmq.XPUB_WELCOME_MSG,
                      ZMQ_STREAM_NOTIFY: zmq.STREAM_NOTIFY,
                      ZMQ_INVERT_MATCHING: zmq.INVERT_MATCHING,
                      ZMQ_HEARTBEAT_IVL: zmq.HEARTBEAT_IVL,
                      ZMQ_HEARTBEAT_TTL: zmq.HEARTBEAT_TTL,
                      ZMQ_HEARTBEAT_TIMEOUT: zmq.HEARTBEAT_TIMEOUT,
                      ZMQ_XPUB_VERBOSER: zmq.XPUB_VERBOSER,
                      ZMQ_CONNECT_TIMEOUT: zmq.CONNECT_TIMEOUT,
                      ZMQ_TCP_MAXRT: zmq.TCP_MAXRT,
                      ZMQ_THREAD_SAFE: zmq.THREAD_SAFE,
                      ZMQ_MULTICAST_MAXTPDU: zmq.MULTICAST_MAXTPDU,
                      ZMQ_VMCI_BUFFER_SIZE: zmq.VMCI_BUFFER_SIZE,
                      ZMQ_VMCI_BUFFER_MIN_SIZE: zmq.VMCI_BUFFER_MIN_SIZE,
                      ZMQ_VMCI_BUFFER_MAX_SIZE: zmq.VMCI_BUFFER_MAX_SIZE,
                      ZMQ_VMCI_CONNECT_TIMEOUT: zmq.VMCI_CONNECT_TIMEOUT,
                      ZMQ_USE_FD: zmq.USE_FD,
                      ZMQ_GSSAPI_PRINCIPAL_NAMETYPE: zmq.GSSAPI_PRINCIPAL_NAMETYPE,
                      ZMQ_GSSAPI_SERVICE_PRINCIPAL_NAMETYPE: zmq.GSSAPI_SERVICE_PRINCIPAL_NAMETYPE,
                      ZMQ_BINDTODEVICE: zmq.BINDTODEVICE,
                      ZMQ_IDENTITY: zmq.IDENTITY,
                      ZMQ_CONNECT_RID: zmq.CONNECT_RID,
                      ZMQ_TCP_ACCEPT_FILTER: zmq.TCP_ACCEPT_FILTER,
                      ZMQ_IPC_FILTER_PID: zmq.IPC_FILTER_PID,
                      ZMQ_IPC_FILTER_UID: zmq.IPC_FILTER_UID,
                      ZMQ_IPC_FILTER_GID: zmq.IPC_FILTER_GID,
                      ZMQ_IPV4ONLY: zmq.IPV4ONLY,
                      ZMQ_DELAY_ATTACH_ON_CONNECT: zmq.DELAY_ATTACH_ON_CONNECT,
                      ZMQ_FAIL_UNROUTABLE: zmq.FAIL_UNROUTABLE,
                      ZMQ_ROUTER_BEHAVIOR: zmq.ROUTER_BEHAVIOR,
                      ZMQ_ZAP_ENFORCE_DOMAIN: zmq.ZAP_ENFORCE_DOMAIN,
                      ZMQ_LOOPBACK_FASTPATH: zmq.LOOPBACK_FASTPATH,
                      ZMQ_METADATA: zmq.METADATA,
                      ZMQ_MULTICAST_LOOP: zmq.MULTICAST_LOOP,
                      ZMQ_ROUTER_NOTIFY: zmq.ROUTER_NOTIFY,
                      ZMQ_XPUB_MANUAL_LAST_VALUE: zmq.XPUB_MANUAL_LAST_VALUE,
                      ZMQ_SOCKS_USERNAME: zmq.SOCKS_USERNAME,
                      ZMQ_SOCKS_PASSWORD: zmq.SOCKS_PASSWORD,
                      ZMQ_IN_BATCH_SIZE: zmq.IN_BATCH_SIZE,
                      ZMQ_OUT_BATCH_SIZE: zmq.OUT_BATCH_SIZE,
                      ZMQ_WSS_KEY_PEM: zmq.WSS_KEY_PEM,
                      ZMQ_WSS_CERT_PEM: zmq.WSS_CERT_PEM,
                      ZMQ_WSS_TRUST_PEM: zmq.WSS_TRUST_PEM,
                      ZMQ_WSS_HOSTNAME: zmq.WSS_HOSTNAME,
                      ZMQ_WSS_TRUST_SYSTEM: zmq.WSS_TRUST_SYSTEM,
                      ZMQ_ONLY_FIRST_SUBSCRIBE: zmq.ONLY_FIRST_SUBSCRIBE,
                      ZMQ_RECONNECT_STOP: zmq.RECONNECT_STOP,
                      ZMQ_HELLO_MSG: zmq.HELLO_MSG,
                      ZMQ_DISCONNECT_MSG: zmq.DISCONNECT_MSG,
                      ZMQ_PRIORITY: zmq.PRIORITY
                     }



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import stat
from abc import ABC,                                \
                abstractmethod

from galaxy.net.net import Client,                  \
                           AsyncClient,             \
                           Server,                  \
                           AsyncServer
from galaxy.net.ssh.transport import SSHTransportFactory
from galaxy.perfo.decorator import timed,                               \
                                   async_timed


class SCPClient(Client, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SCPClient, self).__init__()

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    def scp(self, path: str) -> None:
        raise NotImplementedError("Should implement scp()")

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    def __repr__(self) -> str:
        return "<SFTPClient(id='{}')>".format(self.id)


class ParamikoSCPClient(SCPClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ParamikoSCPClient, self).__init__()
        self.transport_fact: SSHTransportFactory | None = None
        self.client = None

    @timed
    def _load(self) -> None:
        super(ParamikoSCPClient, self)._load()
        if self.transport_fact is not None:
            self.transport_fact._load()

    def _connect(self) -> None:
        if self.client is None:
            raise

    def scp(self, path: str) -> None:
        return self.client.chdir(path)

    def _close(self) -> None:
        return self.client.close()

    def __repr__(self) -> str:
        return "<SCPClient(id='{}')>".format(self.id)


class SCPAsyncClient(AsyncClient, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SCPAsyncClient, self).__init__()

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    async def scp(self, path: str) -> None:
        raise NotImplementedError("Should implement scp()")

    @abstractmethod
    async def _close(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<SFTPAsyncClient(id='{}')>".format(self.id)


class SCPServer(Server, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SCPServer, self).__init__()

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SFTPServer(id='{}')>".format(self.id)


class SCPAsyncServer(AsyncServer, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SCPAsyncServer, self).__init__()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SCPAsyncServer(id='{}')>".format(self.id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                            \
                abstractmethod
from typing import Any

from galaxy.utils.base import Component,                        \
                              Configurable


class SessionFactory(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)

    @abstractmethod
    def create(self, client: Any) -> Any:
        raise NotImplementedError("Should implement create()")


class AsyncSessionFactory(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)

    @abstractmethod
    async def create(self, client: Any) -> Any:
        raise NotImplementedError("Should implement create()")



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                    \
                abstractmethod
from asyncio import Future,                             \
                    Queue
from aiozmq import ZmqProtocol
from aiozmq.interface import ZmqTransport
from typing import Any

from galaxy.utils.base import Component
from galaxy.service.service import LogService,          \
                                   LogAsyncService


class ZmqProtocolFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqProtocolFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, on_close: Future, queue: Queue | None = None) -> Any:
        raise NotImplementedError("Should implement create()")


class AioZmqProtocol(ZmqProtocol):
    """
    classdocs
    """

    def __init__(self, on_close: Future, log: LogService | LogAsyncService, queue: Queue | None = None) -> None:
        """
        Constructor
        """
        super(AioZmqProtocol, self).__init__()
        self.transport: ZmqTransport | None = None
        self.on_close: Future = on_close
        self.queue: Queue | None = queue
        self.log = log

    def connection_made(self, transport: ZmqTransport) -> None:
        self.transport = transport

    def msg_received(self, msg: bytes) -> None:
        if self.queue is not None:
            self.queue.put_nowait(msg)

    def connection_lost(self, ex: Exception | None) -> None:
        self.log.logger.info("The connection has been lost : {}".format(ex))
        self.on_close.set_result(ex)


class AioZmqProtocolFactory(ZmqProtocolFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioZmqProtocolFactory, self).__init__()

    def create(self, on_close: Future, queue: Queue | None = None) -> AioZmqProtocol:
        return AioZmqProtocol(on_close, self.log, queue)


# class HelloWorldProtocol(AioZmqProtocol):
#
#     def __init__(self, loop):
#         """
#         Constructor
#         """
#         self.transport = None
#         self.wait_ready = Future(loop=loop)
#         self.wait_done = Future(loop=loop)
#         self.wait_closed = Future(loop=loop)
#         self.events_received = Queue(loop=loop)
#
#     def connection_made(self, transport):
#         self.transport = transport
#         self.wait_ready.set_result(True)
#
#     def connection_lost(self, exc):
#         self.wait_closed.set_result(exc)
#
#     def msg_received(self, data):
#         # This protocol is used by both the Router and Dealer sockets.
#         # Messages received by the router come prefixed with an 'identity'
#         # and hence contain two frames in this simple test protocol.
#         if len(data) == 2:
#             identity, msg = data
#             if msg == b'Hello':
#                 self.transport.write([identity, b'World'])
#         else:
#             msg = data[0]
#             if msg == b'World':
#                 self.wait_done.set_result(True)
#
#     def event_received(self, event):
#         self.events_received.put_nowait(event)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import json
import logging
from typing import TYPE_CHECKING
from aiohttp.client import ClientSession
from aiohttp.formdata import FormData
from aiohttp.client_exceptions import ClientOSError,                    \
                                      ServerDisconnectedError

from requests.sessions import Session
from requests.models import Request as Request_
from urllib.request import Request
from typing import Any
from tenacity.asyncio import AsyncRetrying
from tenacity.retry import retry_if_not_result,                         \
                           retry_if_result,                             \
                           retry_if_exception_type
from tenacity.stop import stop_after_attempt,                           \
                          stop_after_delay
from tenacity.wait import wait_fixed
from tenacity import RetryError
from tenacity.before import before_log
from tenacity.after import after_log

from galaxy.net.net import Client,                                      \
                           AsyncClient
from galaxy.net.http.request import HTTPRequestBuilder
from galaxy.net.http.response import HTTPResponseFactory
from galaxy.service import constant
from galaxy.utils.base import Component,                                \
                              Configurable
from galaxy.service.service import LogService,                          \
                                   LogAsyncService
from galaxy.error.net import NetHTTPInternalServerError
from galaxy.perfo.decorator import timed,                               \
                                   async_timed

if TYPE_CHECKING:
    from galaxy.net.http.session import HTTPSessionFactory,             \
                                        HTTPAsyncSessionFactory,        \
                                        UrllibHTTPSessionFactory,       \
                                        RequestsHTTPSessionFactory,     \
                                        AioHTTPAsyncSessionFactory


class HTTPClient(Client):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPClient, self).__init__()
        self.reqreps: dict[str, HTTPRequestReply] | None = None
        self.session_fact: HTTPSessionFactory | None = None

    def _load(self) -> None:
        super(HTTPClient, self)._load()
        [reqrep._load() for reqrep in self.reqreps.values()]

    def _connect(self) -> None:
        super(HTTPClient, self)._connect()

    def _close(self) -> None:
        super(HTTPClient, self)._close()

    def _get_formatted_body(self, request: Request) -> str:
        return request.data

    def __repr__(self) -> str:
        return "<HTTPClient(id='{}')>".format(self.id)


class UrllibHTTPClient(HTTPClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(UrllibHTTPClient, self).__init__()
        self.reqreps: dict[str, HTTPRequestReply] | None = None
        self.session_fact: UrllibHTTPSessionFactory | None = None
        self.session: ClientSession | None = None

    @timed
    def _load(self) -> None:
        super(UrllibHTTPClient, self)._load()
        if self.session_fact is not None:
            self.session_fact._load()

    @timed
    def _connect(self) -> None:
        self.session = self.session_fact.create(self)
        if self.session is None:
            raise

    @timed
    def send(self, request: Request, resp_format: str) -> Any:
        self.log.logger.debug("Send the HTTP {} request {} with data {} and headers {}".format(request.method,
                                                                                               request.get_full_url(),
                                                                                               request.data,
                                                                                               request.headers))
        request.headers.update({"Cookie": self.session.cookies})
        resp = self.session.pool_manager.request(request.method.upper(),
                                                 request.get_full_url(),
                                                 body=self._get_formatted_body(request),
                                                 headers=request.headers)
        if resp.status != 200:
            self.close()
            raise NetHTTPInternalServerError(resp.status)
        cookies = resp.headers.get("Set-Cookie")
        if cookies:
            self.session.cookies.update(cookies)
        if resp_format == "json":
            return resp.json()
        else:
            return resp.data

    @timed
    def _close(self) -> None:
        self.session.close()

    def _get_formatted_body(self, request: Request) -> [str, dict[str, str]]:
        if "content_type" in request.headers:
            if request.headers["content_type"] == "application/json":
                return json.dumps(request.data)
            elif request.headers["content_type"] == "application/x-www-form-urlencoded":
                return request.data
        return request.data

    def __repr__(self) -> str:
        return "<UrllibHTTPClient(id='{}')>".format(self.id)


class RequestsHTTPClient(HTTPClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(RequestsHTTPClient, self).__init__()
        self.reqreps: dict[str, HTTPRequestReply] | None = None
        self.session_fact: RequestsHTTPSessionFactory | None = None
        self.session: Session | None = None

    @timed
    def _load(self) -> None:
        super(RequestsHTTPClient, self)._load()
        if self.session_fact is not None:
            self.session_fact._load()

    @timed
    def _connect(self) -> None:
        self.session = self.session_fact.create(self)
        if self.session is None:
            raise

    @timed
    def send(self, request: Request, resp_format: str) -> Any:
        data = self._get_formatted_body(request)
        self.log.logger.debug("Send the HTTP {} request {} with data {} and headers {}".format(request.method,
                                                                                               request.get_full_url(),
                                                                                               data,
                                                                                               request.headers))
        req = Request_(request.method.upper(),
                       request.get_full_url(),
                       data=data,
                       headers=request.headers)
        prepped = self.session.prepare_request(req)
        try:
            resp = self.session.send(prepped)
        except Exception as e:
            print(e)
        if resp.status_code != 200:
            self.close()
            raise NetHTTPInternalServerError(resp.status_code)
        if resp_format == "json":
            return resp.json()
        elif resp_format == "txt":
            return resp.text
        else:
            return resp.content

    @timed
    def _close(self) -> None:
        self.session.close()

    def _get_formatted_body(self, request: Request) -> [str, dict[str, str]]:
        headers = {k.lower().replace("-", "_"): v for k, v in request.headers.items()}
        if "content_type" in headers:
            if headers["content_type"] == "application/json":
                return json.dumps(request.data)
            elif headers["content_type"] == "application/x-www-form-urlencoded":
                return request.data
        return request.data

    def __repr__(self) -> str:
        return "<RequestsHTTPClient(id='{}')>".format(self.id)


class HTTPAsyncClient(AsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPAsyncClient, self).__init__()
        self.reqreps: dict[str, HTTPRequestReply] | None = None
        self.session_fact: HTTPAsyncSessionFactory | None = None

    async def _load(self) -> None:
        await super(HTTPAsyncClient, self)._load()
        [reqrep._load() for reqrep in self.reqreps.values()]

    async def _connect(self) -> None:
        await super(HTTPAsyncClient, self)._connect()

    async def _close(self) -> None:
        await super(HTTPAsyncClient, self)._close()

    def _get_formatted_body(self, request: Request) -> [str, FormData]:
        return request.data

    def __repr__(self) -> str:
        return "<HTTPAsyncClient(id='{}')>".format(self.id)


class AioHTTPAsyncClient(HTTPAsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioHTTPAsyncClient, self).__init__()
        self.reqreps: dict[str, HTTPRequestReply] | None = None
        self.session_fact: AioHTTPAsyncSessionFactory | None = None
        self.session: ClientSession | None = None

    @async_timed
    async def _load(self) -> None:
        await super(AioHTTPAsyncClient, self)._load()
        if self.session_fact is not None:
            self.session_fact._load()

    @async_timed
    async def _connect(self) -> None:
        self.session = await self.session_fact.create(self)
        if self.session is None:
            raise

    @async_timed
    async def _send(self, request: Request, resp_format: str, attempt_nb: int) -> tuple[int, Any]:
        res = None
        data = self._get_formatted_body(request)
        self.log.logger.debug("Send the HTTP {} request {} (Attempt {}) with data {} and headers {}".format(request.method,
                                                                                                            request.get_full_url(),
                                                                                                            attempt_nb,
                                                                                                            data,
                                                                                                            request.headers))
        try:
            async with self.session.request(request.method.lower(),
                                            request.get_full_url(),
                                            data=data,
                                            headers=request.headers) as resp:
                status = resp.status
                self.log.logger.debug("The HTTP {} request {} (Attempt {}) returns with status {}".format(request.method,
                                                                                                          request.get_full_url(),
                                                                                                          attempt_nb,
                                                                                                          status))
                if status == 200:
                    if resp_format == "json":
                        res = await resp.json()
                    elif resp_format == "txt":
                        res = await resp.text()
        except Exception as e:
            self.log.logger.error("The HTTP {} request {} (Attempt {}) fails with the exception '{}'".format(request.method,
                                                                                                             request.get_full_url(),
                                                                                                             attempt_nb,
                                                                                                             str(e)))
            raise e
        return (status, res)

    # Sending HTTP request with asyncio timeout-
    @async_timed
    async def _send_with_timeout(self, request: Request, resp_format: str, attempt_nb: int) -> tuple[int, Any]:
        if "timeout" in self.session_fact.conf:
            timeout = self.session_fact.conf["timeout"].get("total", 120)
            try:
                self.log.logger.debug("Send the HTTP {} request {} (Attempt {}) with a timeout of {} sec".format(request.method,
                                                                                                                 request.get_full_url(),
                                                                                                                 attempt_nb,
                                                                                                                 timeout))
                status, resp = await asyncio.wait_for(self._send(request, resp_format, attempt_nb), timeout)
                # in Python 3.11 :
                #async with asyncio.timeout(timeout):
                #    status, resp = await self._send_request(request, resp_format, attempt_nb)
            except asyncio.TimeoutError as e:
                self.log.logger.error("The HTTP {} request {} (Attempt {}) fails with a timeout of {} sec".format(request.method,
                                                                                                                  request.get_full_url(),
                                                                                                                  attempt_nb,
                                                                                                                  timeout))
                raise e
        else:
            status, resp = await self._send(request, resp_format, attempt_nb)
        return (status, resp)

    @async_timed
    async def send(self, request: Request, resp_format: str) -> Any:
        resp = None
        kwargs = {
                  "retry": (retry_if_not_result(lambda s: s == 200) |
                            retry_if_exception_type(ConnectionResetError) |
                            retry_if_exception_type(asyncio.TimeoutError) |
                            retry_if_exception_type(asyncio.CancelledError) |
                            retry_if_exception_type(ClientOSError) |
                            retry_if_exception_type(ServerDisconnectedError)),
                  "before": before_log(self.log.logger, logging.DEBUG),
                  "after": after_log(self.log.logger, logging.DEBUG)
                 }
        if self.is_connected:
            if "retries" in self.session_fact.conf:
                retries = self.session_fact.conf["retries"].get("total", 1)
                kwargs["stop"] = stop_after_attempt(retries)
            #stop_after_delay does not seem to work :
            #if "timeout" in self.session_fact.conf:
            #    timeout = self.session_fact.conf["timeout"].get("total", 120)
            #    kwargs["stop"] |= stop_after_delay(timeout)
        if "req_interval" in self.session_fact.conf:
            kwargs["wait"] = wait_fixed(self.session_fact.conf["req_interval"])
        try:
            async for attempt in AsyncRetrying(**kwargs):
                with attempt:
                    status, resp = await self._send(request, resp_format, attempt.retry_state.attempt_number)
                if not attempt.retry_state.outcome.failed:
                    attempt.retry_state.set_result(status)
                    if status != 200:
                        self.log.logger.error("The HTTP {} request {} (Attempt {}) fails with status {}".format(request.method,
                                                                                                                request.get_full_url(),
                                                                                                                attempt.retry_state.attempt_number,
                                                                                                                status))
        except RetryError:
            pass
        return resp

    # @async_timed
    # async def send(self, request: Request, resp_format: str) -> Any:
    #     kwargs = {
    #               "retry": retry_if_result(lambda resp: resp is None),
    #               "before": before_log(self.log.logger, logging.DEBUG),
    #               "after": after_log(self.log.logger, logging.DEBUG)
    #              }
    #     if "reconnect_interval" in self.conf:
    #         kwargs["wait"] = wait_fixed(self.conf["reconnect_interval"])
    #     reconnecting = False
    #     try:
    #         async for attempt in AsyncRetrying(**kwargs):
    #             with attempt:
    #                 if reconnecting:
    #                     await self.connect()
    #                     reconnecting = False
    #                 resp = await self._send_with_retries(request, resp_format)
    #                 if resp is None and self.state != constant.STATE_CLOSED:
    #                     if "reconnect_interval" in self.conf:
    #                         self.log.logger.error("The HTTP {} request {} fails : starting to reconnect after {} sec".format(request.method,
    #                                                                                                                          request.get_full_url(),
    #                                                                                                                          self.conf["reconnect_interval"]))
    #                     else:
    #                         self.log.logger.error("The HTTP {} request {} fails : starting to reconnect".format(request.method,
    #                                                                                                             request.get_full_url()))
    #                     await self.close()
    #                     reconnecting = True
    #             if not attempt.retry_state.outcome.failed:
    #                 attempt.retry_state.set_result(resp)
    #     except RetryError:
    #         pass
    #     return resp

    @async_timed
    async def _close(self) -> None:
        await self.session.close()

    def _get_formatted_body(self, request: Request) -> [str, FormData]:
        headers = {k.lower().replace("-", "_"):v for k,v in request.headers.items()}
        if "content_type" in headers:
            if headers["content_type"] == "application/json":
                return json.dumps(request.data)
            elif headers["content_type"] == "application/x-www-form-urlencoded":
                return FormData(request.data, charset="utf-8")
        return request.data

    def __repr__(self) -> str:
        return "<AioHTTPAsyncClient(id='{}')>".format(self.id)


class HTTPRequestReply(Component, Configurable):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self.log: LogService | LogAsyncService | None = None
        self.req_builder: HTTPRequestBuilder | None = None
        self.resp_fact: HTTPResponseFactory | None = None

    def __repr__(self) -> str:
        return "<HTTPRequestReply(id='{}')>".format(self.id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any

from requests.auth import HTTPBasicAuth
from urllib3.poolmanager import PoolManager
from urllib3.util.retry import Retry
from requests.sessions import Session
from requests.adapters import HTTPAdapter
from aiohttp.client import ClientSession,                           \
                           ClientTimeout
from aiohttp.connector import TCPConnector
from aiohttp.helpers import BasicAuth

from galaxy.net.session import SessionFactory,                      \
                               AsyncSessionFactory
from galaxy.net.http.http import HTTPClient,                        \
                                 HTTPAsyncClient
from galaxy.service.service import LogService,                      \
                                   LogAsyncService
from galaxy.net.http.request import HTTPHeaderBuilder


class HTTPSessionFactory(SessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPSessionFactory, self).__init__()
        self.log: LogService | None = None

    def create(self, client: HTTPClient) -> Any:
        pass


class UrllibHTTPSession(object):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self.pool_manager: PoolManager = PoolManager()
        self.cookies: dict[str, str] = {}


class UrllibHTTPSessionFactory(HTTPSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(UrllibHTTPSessionFactory, self).__init__()

    def create(self, client: HTTPClient) -> PoolManager:
        session = UrllibHTTPSession()
        client.session = session
        return session

class RequestsTimeoutSession(Session):
    """
    classdocs
    """
    def __init__(self, connect: float | None = 0, sock_read: float | None = 0) -> None:
        """
        Constructor
        """
        super(RequestsTimeoutSession, self).__init__()
        self.connect = connect
        self.sock_read = sock_read

    def request(self, method, url, **kwargs):
        kwargs.setdefault("timeout", (self.connect, self.sock_read))
        return super(RequestsTimeoutSession, self).request(method, url, **kwargs)


class RequestsHTTPSessionFactory(HTTPSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(RequestsHTTPSessionFactory, self).__init__()

    def create(self, client: HTTPClient) -> Session:
        if "timeout" in self.conf:
            kwargs_timeout = {}
            if "connect" in self.conf["timeout"]:
                kwargs_timeout["connect"] = self.conf["timeout"]["connect"]
            if "sock_read" in self.conf["timeout"]:
                kwargs_timeout["sock_read"] = self.conf["timeout"]["sock_read"]
            session = RequestsTimeoutSession(**kwargs_timeout)
        else:
            session = Session()
        headers = HTTPHeaderBuilder().from_conf(self.conf["header"]).build()
        session.headers.update(headers)
        if "retries" in self.conf:
            kwargs_retries = {}
            if "total" in self.conf["retries"]:
                kwargs_retries["total"] = self.conf["retries"]["total"]
            if "backoff_factor" in self.conf["retries"]:
                kwargs_retries["backoff_factor"] = self.conf["retries"]["backoff_factor"]
            if "status_forcelist" in self.conf["retries"]:
                kwargs_retries["status_forcelist"] = self.conf["retries"]["status_forcelist"]
            if "allowed_methods" in self.conf["retries"]:
                kwargs_retries["allowed_methods"] = frozenset(self.conf["retries"]["allowed_methods"])
            retries = Retry(**kwargs_retries)
            session.mount("https://", HTTPAdapter(max_retries=retries))
            session.mount("http://", HTTPAdapter(max_retries=retries))
        client.session = session
        return session


class HTTPAsyncSessionFactory(AsyncSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPAsyncSessionFactory, self).__init__()
        self.log: LogAsyncService | None = None

    async def create(self, client: HTTPAsyncClient) -> Any:
        pass


class AioHTTPAsyncSessionFactory(HTTPAsyncSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioHTTPAsyncSessionFactory, self).__init__()

    async def create(self, client: HTTPAsyncClient) -> ClientSession:
        headers = HTTPHeaderBuilder().from_conf(self.conf["header"]).build()
        kwargs = {}
        if "timeout" in self.conf:
            kwargs_timeout = {}
            if "total" in self.conf["timeout"]:
                kwargs_timeout["total"] = self.conf["timeout"]["total"]
            if "connect" in self.conf["timeout"]:
                kwargs_timeout["connect"] = self.conf["timeout"]["connect"]
            if "sock_read" in self.conf["timeout"]:
                kwargs_timeout["sock_read"] = self.conf["timeout"]["sock_read"]
            if "sock_connect" in self.conf["timeout"]:
                kwargs_timeout["sock_connect"] = self.conf["timeout"]["sock_connect"]
            timeout = ClientTimeout(**kwargs_timeout)
            kwargs["timeout"] = timeout
        kwargs["headers"] = headers
        kwargs["connector"] = TCPConnector(verify_ssl=self.conf["verify_ssl"])
        if "cred" in self.conf:
            kwargs["auth"] = BasicAuth(login=self.conf["cred"]["username"],
                                       password=self.conf["cred"]["password"],
                                       encoding="utf-8")
        session = ClientSession(**kwargs)
        client.session = session
        return session

    def init_log(self):
        [self.log.add_logger(logger) for logger in ["aiohttp.client",
                                                    "aiohttp.internal"]]



#  Copyright (c) 2024 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import socket

from galaxy.utils.base import Component
from galaxy.service.service import LogService,          \
                                   LogAsyncService


class SocketFactory(Component):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SocketFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    def create(self, host: str, port: int) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        sock.bind((host, port))
        return sock



#  Copyright (c) 2022-2025 Sucden Financial Limited.
#
#  Written by bastien.saltel.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                    \
                abstractmethod
import sys
from typing import Any
from asyncio import Future,                             \
                    Queue
from asyncssh.session import DataType
from asyncssh.session import SSHClientSession
import asyncssh
from asyncssh.connection import SSHClientConnection

from galaxy.utils.base import Component
from galaxy.service.service import LogService,          \
                                   LogAsyncService


class SSHSessionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHSessionFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, on_close: Future, queue: Queue = None) -> Any:
        raise NotImplementedError("Should implement create()")


class AsyncSSHSession(SSHClientSession):
    """
    classdocs
    """

    def __init__(self, on_close: Future, log: LogService | LogAsyncService, queue: Queue | None = None) -> None:
        """
        Constructor
        """
        super(AsyncSSHSession, self).__init__()
        self.transport: SSHClientConnection | None = None
        self.on_close: Future = on_close
        self.queue: Queue | None = queue
        self.log = log

    def data_received(self, data: str, datatype: DataType) -> None:
        if datatype == asyncssh.constants.EXTENDED_DATA_STDERR:
            print(data, end="", file=sys.stderr)
        elif self.queue is not None:
            self.queue.put_nowait(data)

    def connection_lost(self, ex: Exception | None) -> None:
        if ex:
            print("SSH session error: " + str(ex), file=sys.stderr)
        self.on_close.set_result(ex)


class AsyncSSHSessionFactory(SSHSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHSessionFactory, self).__init__()

    def create(self, on_close: Future, queue: Queue | None = None) -> AsyncSSHSession:
        return AsyncSSHSession(on_close, self.log, queue)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                                    \
                abstractmethod

from galaxy.utils.base import Component
from galaxy.service.service import LogService,                          \
                                   LogAsyncService
from galaxy.net import constant
from galaxy.utils.pattern import Builder


class Uri(object):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self.host: str | None = None
        self.port: int = constant.HTTP_PORT
        self.scheme: str = constant.HTTP_SCHEME
        self.path: str | None = None
        self.query: str | None = None
        self.username: str | None = None
        self.password: str | None = None
        self.userinfo: str | None = None
        self.fragment: str | None = None

    def _get_uri(self) -> str:
        uri = ""
        if self.scheme is not None and len(self.scheme) > 0:
            uri += self.scheme + ":"
        if self.host is not None and len(self.host) > 0:
            uri += "//"
            if self.userinfo is not None and len(self.userinfo) > 0:
                uri += "{}@".format(self.username)
            uri += self.host
            if self.port is not None:
                uri += ":{}".format(self.port)
        if self.path is not None and len(self.path) > 0:
            uri += self.path
            if self.query is not None and len(self.query) > 0:
                uri += "?{}".format(self.query)
            if self.fragment is not None and len(self.fragment) > 0:
                uri += "#{}".format(self.fragment)
        return uri

    def __str__(self) -> str:
        return self._get_uri()

    def __repr__(self) -> str:
        return "<Uri(uri='{}')>".format(self._get_uri())

    def __eq__(self, other: "Uri") -> bool:
        return self._get_uri() == other._get_uri()

    def __ne__(self, other: "Uri") -> bool:
        return self._get_uri() != other._get_uri()


class UriPathFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(UriPathFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, **kwargs) -> str:
        raise NotImplementedError("Should implement create()")


class UriQueryFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(UriQueryFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, **kwargs) -> str:
        raise NotImplementedError("Should implement create()")


class UriBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(UriBuilder, self).__init__()
        self._host: str | None = None
        self._port: int | None = None
        self._scheme: str | None = None
        self._path: str | None = None
        self._query: str | None = None
        self._username: str | None = None
        self._password: str | None = None
        self._fragment: str | None = None

    def host(self, host: str) -> "UriBuilder":
        self._host = host
        return self

    def port(self, port: int) -> "UriBuilder":
        self._port = port
        return self

    def scheme(self, scheme: str) -> "UriBuilder":
        self._scheme = scheme
        return self

    def path(self, path: str) -> "UriBuilder":
        self._path = path
        return self

    def query(self, query: str) -> "UriBuilder":
        self._query = query
        return self

    def username(self, username: str) -> "UriBuilder":
        self._username = username
        return self

    def password(self, password: str) -> "UriBuilder":
        self._password = password
        return self

    def fragment(self, fragment: str) -> "UriBuilder":
        self._fragment = fragment
        return self

    def from_conf(self, conf: dict) -> "UriBuilder":
        if "host" in conf:
            self._host = conf["host"]
        if "port" in conf:
            self._port = conf["port"]
        if "scheme" in conf:
            self._scheme = conf["scheme"]
        if "path" in conf:
            self._path = conf["path"]
        if "query" in conf:
            self._query = conf["query"]
        if "username" in conf:
            self._username = conf["username"]
        if "password" in conf:
            self._password = conf["password"]
        if "fragment" in conf:
            self._fragment = conf["fragment"]
        return self

    def build(self) -> Uri:
        uri = Uri()
        uri.host = self._host
        uri.port = self._port
        uri.scheme = self._scheme
        uri.path = self._path
        uri.query = self._query
        uri.username = self._username
        uri.password = self._password
        if uri.username is not None and len(uri.username) > 0:
            if uri.password is not None and len(uri.password) > 0:
                uri.userinfo = "{}:{}".format(self._username, self._password)
            else:
                uri.userinfo = self._username
        else:
            uri.userinfo = None
        uri.fragment = self._fragment
        return uri



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                                    \
                abstractmethod

from galaxy.utils.base import Component
from galaxy.service.service import LogService,                          \
                                   LogAsyncService


class HeaderFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HeaderFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, **kwargs) -> dict[str, str]:
        raise NotImplementedError("Should implement create()")


class HeaderBuilder(ABC):
    """
    classdocs
    """

    @abstractmethod
    def build_headers(self, **kwargs) -> "HeaderBuilder":
        raise NotImplementedError("Should implement build_headers()")

    @abstractmethod
    def build(self) -> dict[str, str]:
        raise NotImplementedError("Should implement build()")


class BodyFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(BodyFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, **kwargs) -> str | dict[str, str]:
        raise NotImplementedError("Should implement create()")



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                    \
                abstractmethod
from typing import Any
from requests.sessions import Session
import httpx
from zeep.transports import Transport,                  \
                            AsyncTransport

from galaxy.utils.base import Component,                \
                              Configurable
from galaxy.service.service import LogService,          \
                                   LogAsyncService
from galaxy.net.auth import HTTPAuthenticationFactory,  \
                            CredentialBuilder


class SOAPTransportFactory(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self) -> Any:
        raise NotImplementedError("Should implement create()")


class ZeepSOAPTransportFactory(SOAPTransportFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZeepSOAPTransportFactory, self).__init__()
        self.auth_fact: HTTPAuthenticationFactory | None = None

    def create(self) -> Transport:
        cred = CredentialBuilder().from_conf(self.conf["cred"]).build()
        auth = self.auth_fact.create(cred)
        session = Session()
        session.auth = auth
        return Transport(session=session)


class ZeepSOAPAsyncTransportFactory(SOAPTransportFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZeepSOAPAsyncTransportFactory, self).__init__()
        self.auth_fact: HTTPAuthenticationFactory | None = None

    def create(self) -> AsyncTransport:
        cred = CredentialBuilder().from_conf(self.conf["cred"]).build()
        auth = self.auth_fact.create(cred)
        if "timeout" in self.conf:
            timeout = httpx.Timeout(self.conf["timeout"]["default"], connect=self.conf["timeout"]["connect"])
            http_client = httpx.Client(auth=auth, timeout=timeout)
            http_asyncclient = httpx.AsyncClient(auth=auth, timeout=timeout)
        else:
            http_client = httpx.Client(auth=auth)
            http_asyncclient = httpx.AsyncClient(auth=auth)
        return AsyncTransport(client=http_asyncclient, wsdl_client=http_client)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                    \
                abstractmethod
from typing import Any
from requests.auth import HTTPBasicAuth
import requests_ntlm
import httpx_ntlm

from galaxy.utils.base import Component,                \
                              Configurable
from galaxy.utils.pattern import Builder
from galaxy.service.log import LogService,              \
                               LogAsyncService


class Credential(object):
    """
    classdocs
    """

    def __init__(self,
                 username: str | None = None,
                 password: str | None = None,
                 domain: str | None = None,
                 private_key: str | None = None,
                 passphrase: str | None = None,
                 tenant_id: str | None = None,
                 client_id: str | None = None,
                 secret: str | None = None) -> None:
        """
        Constructor
        """
        self.username: str | None = username
        self.password: str | None = password
        self.domain: str | None = domain
        self.private_key: str | None = private_key
        self.passphrase: str | None = passphrase
        self.tenant_id: str | None = tenant_id
        self.client_id: str | None = client_id
        self.secret: str | None = secret

    def get_user(self) -> str:
        if self.domain is None:
            return self.username
        return "{}\\{}".format(self.domain, self.username)

    def __str__(self) -> str:
        return self.username

    def __repr__(self) -> str:
        return "<Authentication(username='{}')>".format(self.username)

    def __eq__(self, other: "Credential") -> bool:
        return self.username == other.username and self.domain == other.domain

    def __ne__(self, other: "Credential") -> bool:
        return self.username != other.username or self.domain != other.domain


class CredentialBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(CredentialBuilder, self).__init__()
        self._username: str | None = None
        self._password: str | None = None
        self._domain: str | None = None
        self._private_key: str | None = None
        self._passphrase: str | None = None
        self._tenant_id: str | None = None
        self._client_id: str | None = None
        self._secret: str | None = None

    def username(self, username: str) -> "CredentialBuilder":
        self._username = username
        return self

    def password(self, password: str) -> "CredentialBuilder":
        self._password = password
        return self

    def domain(self, domain: str) -> "CredentialBuilder":
        self._domain = domain
        return self

    def private_key(self, private_key: str) -> "CredentialBuilder":
        self._private_key = private_key
        return self

    def passphrase(self, passphrase: str) -> "CredentialBuilder":
        self._passphrase = passphrase
        return self

    def tenant_id(self, tenant_id: str) -> "CredentialBuilder":
        self._tenant_id = tenant_id
        return self

    def client_id(self, client_id: str) -> "CredentialBuilder":
        self._client_id = client_id
        return self

    def secret(self, secret: str) -> "CredentialBuilder":
        self._secret = secret
        return self

    def from_conf(self, conf: dict) -> "CredentialBuilder":
        if "username" in conf:
            self._username = conf["username"]
        if "password" in conf:
            self._password = conf["password"]
        if "domain" in conf:
            self._domain = conf["domain"]
        if "private_key" in conf:
            self._private_key = conf["private_key"]
        if "passphrase" in conf:
            self._passphrase = conf["passphrase"]
        if "tenant_id" in conf:
            self._tenant_id = conf["tenant_id"]
        if "client_id" in conf:
            self._client_id = conf["client_id"]
        if "secret" in conf:
            self._secret = conf["secret"]
        return self

    def build(self) -> Credential:
        return Credential(self._username,
                          self._password,
                          self._domain,
                          self._private_key,
                          self._passphrase,
                          self._tenant_id,
                          self._client_id,
                          self._secret)


class HTTPAuthenticationFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPAuthenticationFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, cred: Credential) -> Any:
        raise NotImplementedError("Should implement create()")


class HTTPBasicAuthenticationFactory(HTTPAuthenticationFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPBasicAuthenticationFactory, self).__init__()

    def create(self, cred: Credential) -> Any:
        raise NotImplementedError("Should implement create()")


class RequestsHTTPBasicAuthenticationFactory(HTTPBasicAuthenticationFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(RequestsHTTPBasicAuthenticationFactory, self).__init__()

    def create(self, cred: Credential) -> HTTPBasicAuth:
        return HTTPBasicAuth(cred.get_user(), cred.password)


class HTTPNTLMAuthenticationFactory(HTTPAuthenticationFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPNTLMAuthenticationFactory, self).__init__()

    def create(self, cred: Credential) -> Any:
        raise NotImplementedError("Should implement create()")


class RequestsHTTPNTLMAuthenticationFactory(HTTPNTLMAuthenticationFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(RequestsHTTPNTLMAuthenticationFactory, self).__init__()

    def create(self, cred: Credential) -> requests_ntlm.HttpNtlmAuth:
        return requests_ntlm.HttpNtlmAuth(cred.get_user(), cred.password)


class HTTPXHTTPNTLMAuthenticationFactory(HTTPNTLMAuthenticationFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPXHTTPNTLMAuthenticationFactory, self).__init__()

    def create(self, cred: Credential) -> httpx_ntlm.HttpNtlmAuth:
        return httpx_ntlm.HttpNtlmAuth(cred.get_user(), cred.password)
